@@ -1,3 +1,15 @@
+//! The index (lookup table of words) lives here.
+//! The trait [`Provider`] (and [`OccurenceProvider`])
+//! enables multiple types of indices to be defined.
+//!
+//! The only one (for now) is [`Simple`]. That stores a list of all the documents
+//! which contains each word in the input data (e.g. web pages). It then fetches those documents
+//! again and finds occurrences within those.
+//!
+//! ---
+//!
+//! The [`DocumentMap`] makes it performant to get the document ID from name and vice versa.
+//!
 //! # Big O notation reference
 //!
 //! `O(1) < O(log n) < O(log n * log n) < O(sqrt(n)) < O(n) < O(n * log n) < O(n^1.1) < O(n²) < O(n³) < O(2^n)`
@@ -31,50 +43,97 @@ impl Id {
 #[derive(Debug, PartialEq, PartialOrd, Clone)]
 pub struct Occurence {
     start: usize,
-    document_id: Id,
+    pub(crate) document_id: Id,
+    pub(crate) word_id: u32,
     rating: f32,
 }
 impl Occurence {
-    fn new(pos: usize, document_id: Id) -> Self {
+    #[inline]
+    fn new(pos: usize, document_id: Id, word_id: u32) -> Self {
         Self {
             start: pos,
             document_id,
+            word_id,
             rating: 0.0,
         }
     }
     #[must_use]
+    #[inline]
     pub fn start(&self) -> usize {
         self.start
     }
     #[must_use]
+    #[inline]
     pub fn id(&self) -> Id {
         self.document_id
     }
     #[must_use]
+    #[inline]
+    pub fn word_id(&self) -> u32 {
+        self.word_id
+    }
+    #[must_use]
+    #[inline]
     pub fn rating(&self) -> f32 {
         self.rating
     }
+    #[inline]
     pub(crate) fn rating_mut(&mut self) -> &mut f32 {
         &mut self.rating
     }
+    #[inline]
+    pub(crate) fn start_mut(&mut self) -> &mut usize {
+        &mut self.start
+    }
 }
 /// If [`Occurence`] is part of an AND, these can be associated to tell where the other parts of
 /// the AND chain are.
 #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
 pub struct AssociatedOccurrence {
     start: usize,
+    word_id: u32,
+}
+impl From<&crate::Hit> for AssociatedOccurrence {
+    #[inline]
+    fn from(hit: &crate::Hit) -> Self {
+        Self {
+            start: hit.start(),
+            word_id: hit.word_id(),
+        }
+    }
+}
+impl From<&Occurence> for AssociatedOccurrence {
+    #[inline]
+    fn from(occ: &Occurence) -> Self {
+        Self {
+            start: occ.start(),
+            word_id: occ.word_id(),
+        }
+    }
 }
 impl AssociatedOccurrence {
-    pub(crate) fn new(start: usize) -> Self {
-        Self { start }
+    pub(crate) fn new(start: usize, word_id: u32) -> Self {
+        Self { start, word_id }
     }
     /// Get the associated occurrence's start.
     #[must_use]
+    #[inline]
     pub fn start(&self) -> usize {
         self.start
     }
+    /// Get the associated occurrence's word id.
+    #[must_use]
+    #[inline]
+    pub fn word_id(&self) -> u32 {
+        self.word_id
+    }
 }
 
+/// Wrapper for representing `T` as only containing alphanumeric characters.
+///
+/// The [`Eq`] and [`Ord`] implementations discard any eventual non-alphanumeric characters.
+///
+/// Can with benefits be used with [`StrPtr`].
 #[derive(Clone, Hash)]
 #[repr(transparent)]
 pub struct Alphanumeral<T: ?Sized>(T);
@@ -141,11 +200,12 @@ impl<T: AsRef<str>> Display for Alphanumeral<T> {
     }
 }
 
-/// Needed to index [custom struct](Alphanumeral) in maps.
+/// Needed to index a [custom struct](Alphanumeral) in maps.
 /// We have to have the same type, so this acts as both the borrowed and owned.
 pub struct StrPtr {
     s: *const str,
-    drop: bool,
+    /// drop if it's owned
+    owned: bool,
 }
 impl StrPtr {
     /// # Safety
@@ -156,7 +216,7 @@ impl StrPtr {
     #[allow(clippy::inline_always)]
     #[inline(always)]
     unsafe fn new(s: &str, drop: bool) -> Self {
-        Self { s, drop }
+        Self { s, owned: drop }
     }
     /// # Safety
     ///
@@ -187,7 +247,7 @@ impl StrPtr {
 }
 impl Drop for StrPtr {
     fn drop(&mut self) {
-        if self.drop {
+        if self.owned {
             // SAFETY: Upheld by caller.
             unsafe { (self.as_mut()).drop_in_place() }
         }
@@ -316,7 +376,8 @@ impl Default for DocumentMap {
     }
 }
 
-#[derive(Debug)]
+/// [`Eq`] isn't implemented as you'd probably want to check which document it belongs to as well.
+#[derive(Debug, Clone)]
 pub struct WordOccurrence {
     pos: usize,
 }
@@ -332,8 +393,14 @@ impl WordOccurrence {
 pub fn next_char(c: char) -> char {
     let mut int = c as u32;
     let max = char::MAX as u32;
+    // Checking the Rust source, all u32 below [`char::MAX`] and not between the range specified
+    // below are OK, so the loop here shouldn't do anything.
+    // Keeping it just in case of an API change, but since this is unicode, that's highly unlikely.
     loop {
         int += 1;
+        if (0xD800..=0xDFFF).contains(&int) {
+            int = 0xE000;
+        }
         if let Some(c) = char::from_u32(int) {
             return c;
         }
@@ -348,6 +415,9 @@ pub trait Provider<'a> {
     type WordIter: Iterator<Item = &'a Arc<AlphanumRef>> + 'a;
     type WordFilteredIter: Iterator<Item = &'a Arc<AlphanumRef>> + 'a;
 
+    /// The implementation should convert `word` to lower-case and remove any hyphen `-`.
+    ///
+    /// [`Alphanumeral`] can be used for this purpose.
     fn insert_word(&mut self, word: impl AsRef<str>, document: Id);
     fn remove_document(&mut self, document: Id);
     fn contains_word(&self, word: impl AsRef<str>, document: Id) -> bool;
@@ -365,12 +435,14 @@ pub trait Provider<'a> {
 
     /// Only adds words which are alphanumeric.
     fn digest_document(&mut self, id: Id, content: &str) {
-        for (_, word) in SplitNonAlphanumeric::new(content) {
+        for item in SplitNonAlphanumeric::new(content) {
+            let word = item.word();
             if word.is_empty() {
                 continue;
             }
             // Word must be alphanumeric to be added.
-            if word.contains(|c: char| !c.is_alphanumeric()) {
+            // Words with hyphens are also accepted.
+            if word.contains(|c: char| !c.is_alphanumeric() && c != '-') {
                 continue;
             }
 
@@ -548,34 +620,92 @@ impl<'a> Provider<'a> for Simple {
     }
 }
 
+#[derive(Debug, Clone)]
+enum SplitItem<'a> {
+    /// A word was found
+    Word { word: &'a str, index: usize },
+    /// A hyphen-separated word was found.
+    /// This can be treated as **one** word and be added, which improves search results.
+    ///
+    /// Adding this makes query `nextgen` match text `next-gen`.
+    /// For some magical reason, the [`crate::proximity`] already provided query `generation`
+    /// matching for `nextgenerationnext`, even though it's in the middle of a string.
+    Hyphen {
+        /// The word with the hyphen still in it.
+        word: &'a str,
+        /// Index of the start of the word in the original string.
+        index: usize,
+    },
+}
+impl<'a> SplitItem<'a> {
+    #[inline]
+    fn word(&self) -> &'a str {
+        match self {
+            Self::Word { word, index: _ } | Self::Hyphen { word, index: _ } => word,
+        }
+    }
+    #[inline]
+    fn index(&self) -> usize {
+        match self {
+            Self::Word { word: _, index } | Self::Hyphen { word: _, index } => *index,
+        }
+    }
+}
 #[derive(Debug)]
 struct SplitNonAlphanumeric<'a> {
     s: &'a str,
     start: usize,
+    hyphen_start: usize,
+    hyphen_second_word: bool,
 }
 impl<'a> SplitNonAlphanumeric<'a> {
     fn new(s: &'a str) -> Self {
-        Self { s, start: 0 }
+        Self {
+            s,
+            start: 0,
+            hyphen_start: 0,
+            hyphen_second_word: false,
+        }
     }
 }
 impl<'a> Iterator for SplitNonAlphanumeric<'a> {
-    type Item = (usize, &'a str);
+    type Item = SplitItem<'a>;
     fn next(&mut self) -> Option<Self::Item> {
-        let mut iter = self.s.get(self.start..).unwrap().char_indices();
+        let mut iter = self.s[self.start..].char_indices();
         // Not for because we need `iter.next()` below.
         while let Some((pos, c)) = iter.next() {
             if !c.is_alphanumeric() {
+                if self.hyphen_second_word {
+                    let hyphen_word = &self.s[self.hyphen_start..self.start + pos];
+                    self.hyphen_second_word = false;
+                    // When the caller calls `next` again, we start over at the beginning of this
+                    // word, since we didn't change `self.start` before returning.
+                    return Some(SplitItem::Hyphen {
+                        word: hyphen_word,
+                        index: self.hyphen_start,
+                    });
+                }
+                if c == '-' {
+                    self.hyphen_start = self.start;
+                    self.hyphen_second_word = true;
+                }
                 let old_start = self.start;
                 self.start = self
                     .start
                     .saturating_add(iter.next().map_or(c.len_utf8(), |(pos, _)| pos));
-                return Some((old_start, self.s.get(old_start..old_start + pos).unwrap()));
+                return Some(SplitItem::Word {
+                    index: old_start,
+                    word: self.s.get(old_start..old_start + pos).unwrap(),
+                });
             }
         }
         let s = self.s.get(self.start..).unwrap();
         if !s.is_empty() {
             self.start += s.len();
-            return Some((self.start, s));
+            return Some(SplitItem::Word {
+                index: self.start,
+                word: s,
+            });
         }
         None
     }
@@ -611,7 +741,7 @@ pub struct SimpleOccurrencesIter<'a, I> {
     document_contents: &'a HashMap<Id, Arc<String>>,
 
     #[allow(clippy::type_complexity)] // That's not complex.
-    current: Option<(SplitNonAlphanumeric<'a>, Id, &'a AlphanumRef, f32)>,
+    current: Option<(SplitNonAlphanumeric<'a>, Id, &'a AlphanumRef)>,
     current_doc_matched: bool,
 
     missing: &'a Mutex<Vec<(Id, &'a AlphanumRef)>>,
@@ -636,8 +766,11 @@ impl<'a, I: Iterator<Item = (Id, &'a AlphanumRef, f32)>> SimpleOccurrencesIter<'
 impl<'a, I: Iterator<Item = (Id, &'a AlphanumRef, f32)>> Iterator for SimpleOccurrencesIter<'a, I> {
     type Item = Occurence;
     fn next(&mut self) -> Option<Self::Item> {
-        if let Some((words_in_doc, doc_id, searching_word, proximity)) = &mut self.current {
-            for (start, word_in_doc) in words_in_doc {
+        if let Some((words_in_doc, doc_id, searching_word)) = &mut self.current {
+            for item in words_in_doc {
+                let word_in_doc = item.word();
+                let start = item.index();
+
                 // SAFETY: We only use it in this scope.
                 let word_ptr = unsafe { StrPtr::sref(word_in_doc) };
                 let word_ptr = Alphanumeral::new(word_ptr);
@@ -648,8 +781,8 @@ impl<'a, I: Iterator<Item = (Id, &'a AlphanumRef, f32)>> Iterator for SimpleOccu
                 }
                 if let Some(word_proximity) = self.words.get(&word_ptr) {
                     self.current_doc_matched = true;
-                    let rating = ((*proximity - 1.0) * 4.0) + ((word_proximity - 1.0) * 4.0);
-                    let mut occ = Occurence::new(start, *doc_id);
+                    let rating = (word_proximity - 1.0) * 4.0;
+                    let mut occ = Occurence::new(start, *doc_id, 0);
                     *occ.rating_mut() += rating;
                     return Some(occ);
                 }
@@ -664,18 +797,13 @@ impl<'a, I: Iterator<Item = (Id, &'a AlphanumRef, f32)>> Iterator for SimpleOccu
             }
         }
 
-        let (next_doc, word, proximity) = self.iter.next()?;
+        let (next_doc, word, _) = self.iter.next()?;
         let contents = if let Some(c) = self.document_contents.get(&next_doc) {
             c
         } else {
             return self.next();
         };
-        self.current = Some((
-            SplitNonAlphanumeric::new(contents),
-            next_doc,
-            word,
-            proximity,
-        ));
+        self.current = Some((SplitNonAlphanumeric::new(contents), next_doc, word));
         self.next()
     }
     fn size_hint(&self) -> (usize, Option<usize>) {
@@ -691,7 +819,7 @@ pub struct SimpleOccurences<'a> {
 
     missing: Mutex<Vec<(Id, &'a AlphanumRef)>>,
 }
-/// Once this is used, call [`Simple::apply_missing`] to remove the missing entries found during
+/// Once this is used, call [`Self::missing`] and [`MissingOccurrences::apply`] to remove the missing entries found during
 /// the search.
 ///
 /// # Panics
@@ -763,6 +891,12 @@ impl<'a> OccurenceProvider<'a> for SimpleOccurences<'a> {
         }
     }
 }
+/// A list of missing occurrences collected when searching for occurrences using [`SimpleOccurences`].
+///
+/// This is a result of the data being changed and the index ([`Simple`]) not updating, causing it
+/// to think there should be a match when there actually isn't.
+///
+/// Please call [`Self::apply`] on this.
 #[derive(Debug)]
 pub struct MissingOccurrences {
     list: Vec<(Id, AlphanumRef)>,
@@ -834,15 +968,15 @@ Aliquam euismod, justo eu viverra ornare, ex nisi interdum neque, in rutrum nunc
         // Same problem here as above
         assert_eq!(
             occurrences.next(),
-            Some(Occurence::new(0, doc_map.get_id("doc1").unwrap()))
+            Some(Occurence::new(0, doc_map.get_id("doc1").unwrap(), 0))
         );
         assert_eq!(
             occurrences.next(),
-            Some(Occurence::new(875, doc_map.get_id("doc1").unwrap()))
+            Some(Occurence::new(875, doc_map.get_id("doc1").unwrap(), 0))
         );
         assert_eq!(
             occurrences.next(),
-            Some(Occurence::new(0, doc_map.get_id("doc3").unwrap()))
+            Some(Occurence::new(0, doc_map.get_id("doc3").unwrap(), 0))
         );
         assert_eq!(occurrences.next(), None,);
     }
