@@ -1,3 +1,9 @@
+//! Query parsing and resolving.
+//!
+//! This handles AND, OR, and NOT operations, as well as parentheses which are resolved before the
+//! others.
+
+use std::borrow::Borrow;
 use std::collections::BTreeSet;
 use std::fmt::{self, Debug, Display};
 use std::iter::Peekable;
@@ -5,6 +11,7 @@ use std::iter::Peekable;
 use crate::index::{self, AssociatedOccurrence};
 use crate::index::{Occurence, Provider};
 use crate::proximity::ProximateMap;
+use crate::set::ProgressiveInclusion;
 use crate::{proximity, set};
 pub use parse::{parse, Options as ParseOptions};
 
@@ -43,6 +50,7 @@ impl PartialEq for BinaryPart {
 }
 impl Eq for BinaryPart {}
 
+/// A part of a [`Query`].
 #[derive(PartialEq, Eq, Clone)]
 #[must_use]
 pub enum Part {
@@ -122,7 +130,13 @@ impl Part {
     ) -> Box<(dyn FnMut(&'a str) -> Option<Box<dyn Iterator<Item = T> + 'a>> + 'a)> {
         Box::new(f)
     }
-    fn as_doc_iter<'a, 'b, T: Ord + 'a, AndMI: Iterator<Item = T> + 'a>(
+    fn as_doc_iter<
+        'a,
+        'b,
+        T: Ord + 'a,
+        AndMI: Iterator<Item = T> + 'a,
+        OrMI: Iterator<Item = T> + 'a,
+    >(
         &'a self,
         mut iter: impl std::ops::DerefMut<
                 Target = (impl FnMut(&'a str) -> Option<Box<dyn Iterator<Item = T> + 'a>> + ?Sized + 'a),
@@ -135,25 +149,26 @@ impl Part {
             Box<dyn Iterator<Item = T> + 'a>,
             Box<dyn Iterator<Item = T> + 'a>,
         ) -> AndMI,
+        or_merger: &impl Fn(Box<dyn Iterator<Item = T> + 'a>, Box<dyn Iterator<Item = T> + 'a>) -> OrMI,
     ) -> Result<Box<dyn Iterator<Item = T> + 'a>, IterError> {
         let iter = match self {
             Self::And(pair) => match (&pair.left, &pair.right) {
                 (other, Part::Not(not)) | (Part::Not(not), other) => and_not_modify(
-                    other.as_doc_iter(&mut *iter, and_not_modify, and_merger)?,
-                    not.as_doc_iter(&mut *iter, and_not_modify, and_merger)?,
+                    other.as_doc_iter(&mut *iter, and_not_modify, and_merger, or_merger)?,
+                    not.as_doc_iter(&mut *iter, and_not_modify, and_merger, or_merger)?,
                 ),
                 _ => Self::iter_to_box(and_merger(
                     pair.left
-                        .as_doc_iter(&mut *iter, and_not_modify, and_merger)?,
+                        .as_doc_iter(&mut *iter, and_not_modify, and_merger, or_merger)?,
                     pair.right
-                        .as_doc_iter(&mut *iter, and_not_modify, and_merger)?,
+                        .as_doc_iter(&mut *iter, and_not_modify, and_merger, or_merger)?,
                 )),
             },
-            Self::Or(pair) => Self::iter_to_box(set::union(
+            Self::Or(pair) => Self::iter_to_box(or_merger(
                 pair.left
-                    .as_doc_iter(&mut *iter, and_not_modify, and_merger)?,
+                    .as_doc_iter(&mut *iter, and_not_modify, and_merger, or_merger)?,
                 pair.right
-                    .as_doc_iter(&mut *iter, and_not_modify, and_merger)?,
+                    .as_doc_iter(&mut *iter, and_not_modify, and_merger, or_merger)?,
             )),
             Self::Not(_) => return Err(IterError::StrayNot),
             Self::String(s) => {
@@ -187,6 +202,7 @@ impl Display for Part {
     }
 }
 
+/// An iterator of documents which matches a [`Query`].
 #[derive(Debug)]
 pub struct Documents<'a, 'b, P: Provider<'a>> {
     proximate_map: ProximateMap<'b>,
@@ -210,7 +226,7 @@ impl<'a, 'b, P: Provider<'a>> Documents<'a, 'b, P> {
                 Part::fn_to_box(|s| {
                     let list = self.proximate_map.get_panic(s);
                     Some(
-                        proximity::proximate_word_docs(self.provider, list)
+                        crate::proximity::proximate_word_docs(self.provider, list)
                             .map(|item| item.id)
                             .collect::<BTreeSet<_>>()
                             .into_iter(),
@@ -220,6 +236,7 @@ impl<'a, 'b, P: Provider<'a>> Documents<'a, 'b, P> {
             },
             &|i, _| i,
             &set::intersect,
+            &set::union,
         )
     }
     pub fn take_proximate_map(&mut self) -> ProximateMap<'b> {
@@ -227,6 +244,12 @@ impl<'a, 'b, P: Provider<'a>> Documents<'a, 'b, P> {
     }
 }
 
+/// A query.
+///
+/// This is just a root [`Part`].
+///
+/// Use [`mod@parse`] to create one.
+/// You can also use the [`std::str::FromStr`] implementation.
 #[derive(Debug, PartialEq, Eq, Clone)]
 #[must_use]
 pub struct Query {
@@ -274,41 +297,40 @@ impl Query {
         provider: &'a impl index::OccurenceProvider<'a>,
         distance_threshold: usize,
     ) -> Result<impl Iterator<Item = Hit> + 'a, IterError> {
-        #[derive(Debug)]
+        #[derive(Debug, Clone)]
         struct OccurenceEq(Hit);
         impl OccurenceEq {
-            fn new(occ: Occurence) -> Self {
-                Self(Hit {
-                    occurrence: occ,
-                    associated_occurrences: BTreeSet::new(),
-                    closest_not: None,
-                })
-            }
-            fn inner(self) -> Hit {
-                self.0
+            #[inline]
+            fn new(mut occ: Occurence, word_id: u32) -> Self {
+                occ.word_id = word_id;
+                Self(Hit::new(occ))
             }
             #[must_use]
             fn closest(&self, other: &Self) -> (usize, AssociatedOccurrence) {
                 use std::cmp::Ordering;
-                let mut closest = (usize::MAX, AssociatedOccurrence::new(0));
-                let mut a_iter = self.0.associated_occurrences().map(|c| c.start());
+                let mut closest = (usize::MAX, AssociatedOccurrence::new(0, 0));
+                let mut a_iter = self.0.occurrences();
                 // If `associated_occurrences` has len 0, use only start.
                 // Else, it'll have >=2 occurences, since the `.start()` is also taken as part of
                 // the BTreeSet.
-                let mut a = a_iter.next().unwrap_or_else(|| self.0.start());
-                let mut b_iter = other.0.associated_occurrences().map(|c| c.start());
-                let mut b = b_iter.next().unwrap_or_else(|| other.0.start());
+                let mut a = a_iter.next().unwrap_or_else(|| (&self.0).into());
+                let mut a_start = a.start();
+                let mut b_iter = other.0.occurrences();
+                let mut b = b_iter.next().unwrap_or_else(|| (&other.0).into());
+                let mut b_start = b.start();
                 let mut one_completed = false;
                 loop {
-                    let dist = if a > b { a - b } else { b - a };
-                    closest =
-                        std::cmp::min_by((dist, AssociatedOccurrence::new(b)), closest, |a, b| {
-                            a.0.cmp(&b.0)
-                        });
+                    let dist = if a_start > b_start {
+                        a_start - b_start
+                    } else {
+                        b_start - a_start
+                    };
+                    closest = std::cmp::min_by((dist, b), closest, |a, b| a.0.cmp(&b.0));
                     match a.cmp(&b) {
                         Ordering::Less => {
                             if let Some(next) = a_iter.next() {
                                 a = next;
+                                a_start = a.start();
                             } else if one_completed {
                                 break;
                             } else {
@@ -318,10 +340,11 @@ impl Query {
                         // The words are at the same place!
                         //
                         // This is technically not reachable, but the following codes makes sense.
-                        Ordering::Equal => return (0, AssociatedOccurrence::new(b)),
+                        Ordering::Equal => return (0, b),
                         Ordering::Greater => {
                             if let Some(next) = b_iter.next() {
                                 b = next;
+                                b_start = b.start();
                             } else if one_completed {
                                 break;
                             } else {
@@ -334,21 +357,30 @@ impl Query {
             }
         }
         impl PartialEq for OccurenceEq {
+            #[inline]
             fn eq(&self, other: &Self) -> bool {
                 self.0.id().eq(&other.0.id())
             }
         }
         impl Eq for OccurenceEq {}
         impl Ord for OccurenceEq {
+            #[inline]
             fn cmp(&self, other: &Self) -> std::cmp::Ordering {
                 self.0.id().cmp(&other.0.id())
             }
         }
         impl PartialOrd for OccurenceEq {
+            #[inline]
             fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
                 Some(self.cmp(other))
             }
         }
+        impl Borrow<index::Id> for OccurenceEq {
+            #[inline]
+            fn borrow(&self) -> &index::Id {
+                &self.0.occurrence.document_id
+            }
+        }
 
         #[derive(Debug)]
         struct MergeProximate<I>
@@ -369,10 +401,7 @@ impl Query {
         impl<I: Iterator<Item = OccurenceEq>> Iterator for MergeProximate<I> {
             type Item = OccurenceEq;
             fn next(&mut self) -> Option<Self::Item> {
-                let mut next = match self.iter.next() {
-                    Some(next) => next,
-                    None => return None,
-                };
+                let mut next = self.iter.next()?;
 
                 let peeked = match self.iter.peek() {
                     Some(peeked) => peeked,
@@ -399,156 +428,221 @@ impl Query {
             }
         }
 
+        #[inline]
+        fn abs_diff_occurrence(a: &OccurenceEq, b: &OccurenceEq) -> usize {
+            a.0.start().abs_diff(b.0.start())
+        }
+
+        let mut word_id: u32 = 0;
         self.root()
             .as_doc_iter(
                 &mut move |s| {
-                    provider.occurrences_of_word(s).map(|iter| {
+                    // Logic error if we wrap, but we'd need more than 4M words, and it only
+                    // results in marginally worse relevance for search results and context.
+                    word_id = word_id.wrapping_add(1);
+                    provider.occurrences_of_word(s).map(move |iter| {
                         Part::iter_to_box(MergeProximate::new(
-                            iter.map(OccurenceEq::new),
+                            iter.map(
+                                #[inline]
+                                move |hit| OccurenceEq::new(hit, word_id),
+                            ),
                             distance_threshold,
                         ))
                     })
                 },
                 &|and, not| {
-                    fn get_closest_btreeset<'a, T: Ord>(
-                        set: &'a BTreeSet<T>,
-                        key: &T,
-                    ) -> (Option<&'a T>, Option<&'a T>) {
-                        (set.range(..key).next_back(), set.range(key..).next())
-                    }
-                    /// Returns [`None`] when `set` is empty.
-                    fn get_before_closest_btreeset<'a>(
-                        set: &'a BTreeSet<Hit>,
-                        key: &Hit,
-                    ) -> Option<&'a Hit> {
-                        let (before, after) = get_closest_btreeset(set, key);
-                        match (before, after) {
-                            (Some(before), Some(after)) => {
-                                if after.id() == key.id() {
-                                    Some(after)
-                                } else {
-                                    Some(before)
-                                }
+                    Part::iter_to_box(
+                        crate::set::progressive(
+                            and,
+                            not,
+                            #[inline]
+                            |and, not| and.0.start().cmp(&not.0.start()),
+                            #[inline]
+                            |a, b| (*a).cmp(b),
+                            Some(abs_diff_occurrence),
+                        )
+                        .filter_map(|inclusion| match inclusion {
+                            ProgressiveInclusion::Left(and) => Some(and),
+                            ProgressiveInclusion::Right(_not) => None,
+                            ProgressiveInclusion::Both(mut and, not) => {
+                                let closest = and.closest(&OccurenceEq::new(not.0.occurrence, 0));
+                                // Don't really care about precision.
+                                #[allow(clippy::cast_precision_loss)]
+                                let rating_decrease = 1.0 / (0.0001 * closest.0 as f32 + 0.025);
+                                *and.0.rating_mut() -= rating_decrease;
+                                and.0.closest_not = Some(closest.1);
+                                Some(and)
                             }
-                            (Some(before), None) => Some(before),
-                            (None, Some(after)) => Some(after),
-                            (None, None) => None,
-                        }
-                    }
-
-                    let not: BTreeSet<_> = not.map(OccurenceEq::inner).collect();
-
-                    if not.is_empty() {
-                        return and;
-                    }
-
-                    // `TODO`: Use iterators here again.
-                    // This requires a `iter_set::classify` function which keeps the last `b` if
-                    // there isn't a new one that's closer.
-                    //
-                    // This is because the NOT needs to persist while the ANDs go by.
-                    Part::iter_to_box(and.map(move |mut and| {
-                        // UNWRAP: It's not empty.
-                        let not = get_before_closest_btreeset(&not, &and.0).unwrap();
-                        if not.id() == and.0.id() {
-                            let closest = and.closest(&OccurenceEq(Hit {
-                                occurrence: not.occurrence.clone(),
-                                associated_occurrences: BTreeSet::new(),
-                                closest_not: None,
-                            }));
-                            // Don't really care about precision.
-                            #[allow(clippy::cast_precision_loss)]
-                            let rating_decrease = 1.0 / (0.0001 * closest.0 as f32 + 0.025);
-                            *and.0.rating_mut() -= rating_decrease;
-                            and.0.closest_not = Some(closest.1);
-                        }
-                        and
-                    }))
+                        }),
+                    )
                 },
+                // AND merger
                 &|left, right| {
-                    iter_set::classify(left, right).filter_map(|inclusion| {
-                        if let iter_set::Inclusion::Both(mut a, b) = inclusion {
+                    crate::set::progressive(
+                        left,
+                        right,
+                        #[inline]
+                        |a, b| a.0.start().cmp(&b.0.start()),
+                        // Compares IDs
+                        #[inline]
+                        |a, b| (*a).cmp(b),
+                        None,
+                    )
+                    .filter_map(|inclusion| match inclusion {
+                        ProgressiveInclusion::Both(mut a, b) => {
                             a.0.merge(&b.0);
                             Some(a)
-                        } else {
-                            None
                         }
+                        _ => None,
+                    })
+                },
+                // OR merger
+                &|left, right| {
+                    crate::set::progressive(
+                        left,
+                        right,
+                        #[inline]
+                        |a, b| a.0.start().cmp(&b.0.start()),
+                        // Compares IDs
+                        #[inline]
+                        |a, b| (*a).cmp(b),
+                        None,
+                    )
+                    .map(|inclusion| match inclusion {
+                        ProgressiveInclusion::Both(mut a, b) => {
+                            a.0.merge(&b.0);
+                            a
+                        }
+                        ProgressiveInclusion::Left(a) | ProgressiveInclusion::Right(a) => a,
                     })
                 },
             )
-            .map(|iter| iter.map(OccurenceEq::inner))
             .map(|iter| {
-                iter.map(|mut occurrence| {
-                    let mut closest = usize::MAX;
-                    let mut iter = occurrence.associated_occurrences();
-                    let mut last = iter.next().unwrap_or_else(|| AssociatedOccurrence::new(0));
-                    for occ in iter {
-                        let dist = occ.start() - last.start();
-                        closest = std::cmp::min(dist, closest);
+                iter.map(|occ| {
+                    let mut occ = occ.0;
+                    let (mut closest, mut closest_index) = (usize::MAX, 0);
+                    let mut iter = occ.occurrences();
+                    let mut last = iter
+                        .next()
+                        .unwrap_or_else(|| AssociatedOccurrence::new(0, 0));
+                    for (idx, occ) in iter.enumerate() {
+                        if last.word_id() != occ.word_id() {
+                            let dist = occ.start() - last.start();
+                            if dist < closest {
+                                closest_index = idx + 1;
+                            }
+                            closest = std::cmp::min(dist, closest);
+                        }
                         last = occ;
                     }
                     if closest < usize::MAX {
                         // Don't really care about precision.
                         #[allow(clippy::cast_precision_loss)]
                         let rating_increase = 1.0 / (0.001 * closest as f32 + 0.1);
-                        *occurrence.rating_mut() += rating_increase;
+                        *occ.rating_mut() += rating_increase;
+
+                        // replace `occurrence.start()` with the closest match.
+                        // This requires removing the closest from the set and setting as "main"
+                        // and inserting "main" to the set.
+                        if closest_index != 0 {
+                            let closest = occ.occurrences().nth(closest_index - 1).unwrap();
+                            occ.occurrences.remove(&closest);
+                            occ.occurrences.insert((&occ.occurrence).into());
+                            *occ.occurrence.start_mut() = closest.start();
+                        }
                     }
 
-                    occurrence
+                    occ
                 })
             })
     }
 }
 
+/// An occurrence returned from [`Query`].
 #[derive(Debug, Clone)]
 pub struct Hit {
     occurrence: Occurence,
 
-    associated_occurrences: BTreeSet<AssociatedOccurrence>,
+    /// [`Self::occurrence`] (if [`Self::merged`]) and all associated occurrences.
+    occurrences: BTreeSet<AssociatedOccurrence>,
+    /// If we have been merged. Used to tell if `self.occurrence.start()` is in the occurrences
+    /// BTree. We don't do this by default to avoid an allocation.
+    merged: bool,
 
     closest_not: Option<AssociatedOccurrence>,
 }
 impl Hit {
+    #[inline]
+    fn new(occurrence: Occurence) -> Self {
+        Self {
+            occurrence,
+            occurrences: BTreeSet::new(),
+            merged: false,
+            closest_not: None,
+        }
+    }
     /// Get a reference to the hit's document id.
     #[must_use]
+    #[inline]
     pub fn id(&self) -> index::Id {
         self.occurrence.id()
     }
+    /// Get a reference to the hit's word id.
+    #[must_use]
+    #[inline]
+    pub fn word_id(&self) -> u32 {
+        self.occurrence.word_id()
+    }
     /// Get a reference to the hit's document id.
     ///
-    /// This might be any of the [`Self::associated_occurrences`] if that returns any items.
+    /// This might be any of the [`Self::occurrences`] if that returns any items.
     #[must_use]
+    #[inline]
     pub fn start(&self) -> usize {
         self.occurrence.start()
     }
     /// Get a reference to the hit's occurrence.
     #[must_use]
-    pub fn occurence(&self) -> &Occurence {
+    #[inline]
+    pub fn occurrence(&self) -> &Occurence {
         &self.occurrence
     }
     /// Get the hit's rating.
     #[must_use]
+    #[inline]
     pub fn rating(&self) -> f32 {
         self.occurrence.rating()
     }
+    #[inline]
     fn rating_mut(&mut self) -> &mut f32 {
         self.occurrence.rating_mut()
     }
-    pub fn associated_occurrences(&self) -> impl Iterator<Item = AssociatedOccurrence> + '_ {
-        self.associated_occurrences.iter().copied()
+    /// [`Self::occurrence`] and all associated occurrences.
+    #[inline]
+    pub fn occurrences(&self) -> impl Iterator<Item = AssociatedOccurrence> + '_ {
+        let first = if self.merged {
+            None
+        } else {
+            Some(index::AssociatedOccurrence::new(
+                self.start(),
+                self.occurrence.word_id(),
+            ))
+        };
+        first.into_iter().chain(self.occurrences.iter().copied())
     }
 
-    /// Must have the same [document id](id()).
+    /// Merges two [`Hit`]s, by adding the [`Hit::occurrences`] to our set.
+    ///
+    /// Must have the same [document id](Self::id()).
     fn merge(&mut self, other: &Self) {
-        if self.associated_occurrences.is_empty() {
-            self.associated_occurrences
-                .insert(index::AssociatedOccurrence::new(self.start()));
+        if self.occurrences.is_empty() {
+            self.occurrences.insert((&*self).into());
         }
-        for other_occ in other.associated_occurrences() {
-            self.associated_occurrences.insert(other_occ);
+        for other_occ in other.occurrences() {
+            self.occurrences.insert(other_occ);
         }
-        self.associated_occurrences
-            .insert(index::AssociatedOccurrence::new(other.start()));
+        self.occurrences.insert(other.into());
+        self.merged = true;
     }
 }
 impl PartialEq for Hit {
@@ -577,6 +671,7 @@ pub enum IterError {
     StrayNot,
 }
 
+/// [`Query`] parsing.
 pub mod parse {
     use std::fmt::{self, Debug, Display};
     use std::str::FromStr;
@@ -632,6 +727,7 @@ pub mod parse {
         OperationIsBinary,
     }
 
+    /// The parser. This is public to allow you to change it's state when making a [`Rule`].
     #[derive(Debug)]
     pub struct Parser {
         sub: Option<Box<Parser>>,
@@ -758,7 +854,15 @@ pub mod parse {
             }
             Ok(1)
         }
-        fn finish_part(&mut self, op: Option<Op>, right: Part) {
+        fn finish_part(&mut self, op: Option<Op>, mut right: Part) {
+            if let Some(marker) = self.string_marker.take() {
+                match marker {
+                    Op::Not => right = Part::not(right),
+                    Op::And | Op::Or => {
+                        unreachable!("In `set_string_marker`, we check for binary.")
+                    }
+                }
+            }
             if let Some(op) = op {
                 self.left = Some(self.finish_op(op, right));
             } else {
@@ -864,6 +968,8 @@ pub mod parse {
         c.is_ascii_whitespace() || c == '\u{a0}'
     }
 
+    /// Options to [`parse`]. Here, you can add your own formatting rules, on top of the default
+    /// rules.
     #[derive(Debug)]
     #[must_use]
     pub struct Options {
@@ -1139,6 +1245,14 @@ mod tests {
         );
     }
     #[test]
+    fn parse_parentheses_and_not() {
+        let p = s("icelk -(agde or kvarn)");
+        assert_eq!(p, Part::and("icelk", Part::not(Part::or("kvarn", "agde"))));
+
+        let p = s("icelk - (agde or kvarn)");
+        assert_eq!(p, Part::and("icelk", Part::not(Part::or("kvarn", "agde"))));
+    }
+    #[test]
     fn parse_not() {
         let p = s("not");
         assert_eq!(p, Part::s("not"));
